package com.sun.xml.internal.bind.v2.runtime;

import java.util.Collections;
import com.sun.xml.internal.bind.v2.model.annotation.RuntimeInlineAnnotationReader;
import org.xml.sax.SAXException;
import com.sun.xml.internal.bind.v2.runtime.property.AttributeProperty;
import com.sun.xml.internal.bind.util.Which;
import java.util.List;
import com.sun.xml.internal.bind.v2.runtime.property.Property;
import com.sun.xml.internal.bind.api.AccessorException;
import com.sun.xml.internal.bind.v2.runtime.reflect.Accessor;
import com.sun.xml.internal.bind.api.RawAccessor;
import com.sun.xml.internal.bind.api.BridgeContext;
import com.sun.xml.internal.bind.unmarshaller.DOMScanner;
import org.w3c.dom.Node;
import javax.xml.bind.Binder;
import com.sun.xml.internal.bind.v2.schemagen.XmlSchemaGenerator;
import java.io.IOException;
import com.sun.xml.internal.bind.api.ErrorListener;
import org.xml.sax.SAXParseException;
import javax.xml.bind.SchemaOutputResolver;
import com.sun.xml.internal.txw2.output.ResultFactory;
import javax.xml.transform.Result;
import com.sun.xml.internal.bind.v2.model.core.NonElement;
import javax.xml.bind.JAXBIntrospector;
import javax.xml.bind.Validator;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallerImpl;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.TransformerConfigurationException;
import com.sun.xml.internal.bind.v2.util.XmlFactory;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.Transformer;
import java.util.TreeSet;
import com.sun.xml.internal.bind.v2.util.EditDistance;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Loader;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.TagName;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallingContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeLeafInfo;
import com.sun.xml.internal.bind.v2.model.core.Ref;
import com.sun.xml.internal.bind.v2.model.impl.RuntimeModelBuilder;
import javax.xml.bind.JAXBException;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.lang.reflect.Type;
import com.sun.xml.internal.bind.v2.model.nav.Navigator;
import java.util.Iterator;
import java.util.Collection;
import javax.xml.bind.annotation.XmlAttachmentRef;
import com.sun.xml.internal.bind.v2.model.core.Adapter;
import javax.xml.bind.annotation.XmlList;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;
import com.sun.xml.internal.bind.api.CompositeStructure;
import javax.xml.bind.JAXBElement;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeElementInfo;
import java.util.Arrays;
import java.util.HashSet;
import javax.xml.bind.annotation.XmlSchema;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeClassInfo;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeArrayInfo;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeEnumLeafInfo;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeBuiltinLeafInfo;
import com.sun.xml.internal.bind.v2.model.impl.RuntimeBuiltinLeafInfoImpl;
import java.util.LinkedHashMap;
import java.util.Comparator;
import com.sun.xml.internal.bind.v2.runtime.output.Encoded;
import javax.xml.bind.annotation.XmlNs;
import java.util.Set;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.bind.v2.model.annotation.RuntimeAnnotationReader;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeInfoSet;
import java.lang.ref.WeakReference;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.Marshaller;
import com.sun.istack.internal.Pool;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeTypeInfo;
import javax.xml.namespace.QName;
import java.util.HashMap;
import com.sun.xml.internal.bind.v2.util.QNameMap;
import javax.xml.parsers.DocumentBuilder;
import com.sun.xml.internal.bind.api.Bridge;
import com.sun.xml.internal.bind.api.TypeReference;
import java.util.Map;
import com.sun.xml.internal.bind.api.JAXBRIContext;

public final class JAXBContextImpl extends JAXBRIContext
{
    private final Map<TypeReference, Bridge> bridges;
    private static DocumentBuilder db;
    private final QNameMap<JaxBeanInfo> rootMap;
    private final HashMap<QName, JaxBeanInfo> typeMap;
    private final Map<Class, JaxBeanInfo> beanInfoMap;
    protected Map<RuntimeTypeInfo, JaxBeanInfo> beanInfos;
    private final Map<Class, Map<QName, ElementBeanInfoImpl>> elements;
    public final Pool<Marshaller> marshallerPool;
    public final Pool<Unmarshaller> unmarshallerPool;
    public NameBuilder nameBuilder;
    public final NameList nameList;
    private final String defaultNsUri;
    private final Class[] classes;
    protected final boolean c14nSupport;
    public final boolean xmlAccessorFactorySupport;
    public final boolean allNillable;
    public final boolean retainPropertyInfo;
    public final boolean supressAccessorWarnings;
    public final boolean improvedXsiTypeHandling;
    public final boolean disableSecurityProcessing;
    private WeakReference<RuntimeTypeInfoSet> typeInfoSetCache;
    @NotNull
    private RuntimeAnnotationReader annotationReader;
    private boolean hasSwaRef;
    @NotNull
    private final Map<Class, Class> subclassReplacements;
    public final boolean fastBoot;
    private Set<XmlNs> xmlNsSet;
    private Encoded[] utf8nameTable;
    private static final Comparator<QName> QNAME_COMPARATOR;
    
    public Set<XmlNs> getXmlNsSet() {
        return this.xmlNsSet;
    }
    
    private JAXBContextImpl(JAXBContextImpl.JAXBContextBuilder builder) throws JAXBException {
        this.bridges = new LinkedHashMap();
        this.rootMap = new QNameMap();
        this.typeMap = new HashMap();
        this.beanInfoMap = new LinkedHashMap();
        this.beanInfos = new LinkedHashMap();
        this.elements = new LinkedHashMap();
        this.marshallerPool = new Pool.Impl<Marshaller>() {
            @NotNull
            protected Marshaller create() {
                return JAXBContextImpl.this.createMarshaller();
            }
        };
        this.unmarshallerPool = new Pool.Impl<Unmarshaller>() {
            @NotNull
            protected Unmarshaller create() {
                return JAXBContextImpl.this.createUnmarshaller();
            }
        };
        this.nameBuilder = new NameBuilder();
        this.xmlNsSet = null;
        this.defaultNsUri = builder.defaultNsUri;
        this.retainPropertyInfo = builder.retainPropertyInfo;
        this.annotationReader = builder.annotationReader;
        this.subclassReplacements = builder.subclassReplacements;
        this.c14nSupport = builder.c14nSupport;
        this.classes = builder.classes;
        this.xmlAccessorFactorySupport = builder.xmlAccessorFactorySupport;
        this.allNillable = builder.allNillable;
        this.supressAccessorWarnings = builder.supressAccessorWarnings;
        this.improvedXsiTypeHandling = builder.improvedXsiTypeHandling;
        this.disableSecurityProcessing = builder.disableSecurityProcessing;
        Collection<TypeReference> typeRefs = builder.typeRefs;
        boolean fastB;
        try {
            fastB = Boolean.getBoolean(JAXBContextImpl.class.getName() + ".fastBoot");
        }
        catch (SecurityException e) {
            fastB = false;
        }
        this.fastBoot = fastB;
        RuntimeTypeInfoSet typeSet = this.getTypeInfoSet();
        this.elements.put(null, new LinkedHashMap());
        for (RuntimeBuiltinLeafInfo leaf : RuntimeBuiltinLeafInfoImpl.builtinBeanInfos) {
            LeafBeanInfoImpl<?> bi = new LeafBeanInfoImpl(this, leaf);
            this.beanInfoMap.put(leaf.getClazz(), bi);
            for (QName t : bi.getTypeNames()) {
                this.typeMap.put(t, bi);
            }
        }
        for (RuntimeEnumLeafInfo e2 : typeSet.enums().values()) {
            JaxBeanInfo<?> bi2 = this.getOrCreate(e2);
            for (QName qn : bi2.getTypeNames()) {
                this.typeMap.put(qn, bi2);
            }
            if (e2.isElement()) {
                this.rootMap.put(e2.getElementName(), bi2);
            }
            else {
                continue;
            }
        }
        for (RuntimeArrayInfo a : typeSet.arrays().values()) {
            JaxBeanInfo<?> ai = this.getOrCreate(a);
            for (QName qn : ai.getTypeNames()) {
                this.typeMap.put(qn, ai);
            }
        }
        for (Map.Entry<Class, ? extends RuntimeClassInfo> e3 : typeSet.beans().entrySet()) {
            ClassBeanInfoImpl<?> bi3 = this.getOrCreate((RuntimeClassInfo)e3.getValue());
            XmlSchema xs = (XmlSchema)this.annotationReader.getPackageAnnotation(XmlSchema.class, e3.getKey(), null);
            if (xs != null && xs.xmlns() != null && xs.xmlns().length > 0) {
                if (this.xmlNsSet == null) {
                    this.xmlNsSet = new HashSet();
                }
                this.xmlNsSet.addAll(Arrays.asList(xs.xmlns()));
            }
            if (bi3.isElement()) {
                this.rootMap.put(((RuntimeClassInfo)e3.getValue()).getElementName(), bi3);
            }
            for (QName qn2 : bi3.getTypeNames()) {
                this.typeMap.put(qn2, bi3);
            }
        }
        for (RuntimeElementInfo n : typeSet.getAllElements()) {
            ElementBeanInfoImpl bi4 = this.getOrCreate(n);
            if (n.getScope() == null) {
                this.rootMap.put(n.getElementName(), bi4);
            }
            RuntimeClassInfo scope = n.getScope();
            Class<Object> scopeClazz = (scope == null) ? null : ((Class)scope.getClazz());
            Map<QName, ElementBeanInfoImpl> m = (Map)this.elements.get(scopeClazz);
            if (m == null) {
                m = new LinkedHashMap();
                this.elements.put(scopeClazz, m);
            }
            m.put(n.getElementName(), bi4);
        }
        this.beanInfoMap.put(JAXBElement.class, new ElementBeanInfoImpl(this));
        this.beanInfoMap.put(CompositeStructure.class, new CompositeStructureBeanInfo(this));
        this.getOrCreate(typeSet.getAnyTypeInfo());
        for (JaxBeanInfo bi5 : this.beanInfos.values()) {
            bi5.link(this);
        }
        for (Map.Entry<Class, Class> e4 : RuntimeUtil.primitiveToBox.entrySet()) {
            this.beanInfoMap.put(e4.getKey(), this.beanInfoMap.get(e4.getValue()));
        }
        Navigator<Type, Class, Field, Method> nav = typeSet.getNavigator();
        for (TypeReference tr : typeRefs) {
            XmlJavaTypeAdapter xjta = (XmlJavaTypeAdapter)tr.get(XmlJavaTypeAdapter.class);
            Adapter<Type, Class> a2 = null;
            XmlList xl = (XmlList)tr.get(XmlList.class);
            Class<T> erasedType = (Class)nav.erasure(tr.type);
            if (xjta != null) {
                a2 = new Adapter(xjta.value(), nav);
            }
            if (tr.get(XmlAttachmentRef.class) != null) {
                a2 = new Adapter(SwaRefAdapter.class, nav);
                this.hasSwaRef = true;
            }
            if (a2 != null) {
                erasedType = (Class)nav.erasure(a2.defaultType);
            }
            Name name = this.nameBuilder.createElementName(tr.tagName);
            InternalBridge<Object> bridge;
            if (xl == null) {
                bridge = new BridgeImpl(this, name, this.getBeanInfo(erasedType, true), tr);
            }
            else {
                bridge = new BridgeImpl(this, name, new ValueListBeanInfoImpl(this, erasedType), tr);
            }
            if (a2 != null) {
                bridge = new BridgeAdapter(bridge, (Class)a2.adapterType);
            }
            this.bridges.put(tr, bridge);
        }
        this.nameList = this.nameBuilder.conclude();
        for (JaxBeanInfo bi6 : this.beanInfos.values()) {
            bi6.wrapUp();
        }
        this.nameBuilder = null;
        this.beanInfos = null;
    }
    
    public boolean hasSwaRef() {
        return this.hasSwaRef;
    }
    
    public RuntimeTypeInfoSet getRuntimeTypeInfoSet() {
        try {
            return this.getTypeInfoSet();
        }
        catch (IllegalAnnotationsException e) {
            throw new AssertionError(e);
        }
    }
    
    public RuntimeTypeInfoSet getTypeInfoSet() throws IllegalAnnotationsException {
        if (this.typeInfoSetCache != null) {
            RuntimeTypeInfoSet r = (RuntimeTypeInfoSet)this.typeInfoSetCache.get();
            if (r != null) {
                return r;
            }
        }
        RuntimeModelBuilder builder = new RuntimeModelBuilder(this, this.annotationReader, this.subclassReplacements, this.defaultNsUri);
        IllegalAnnotationsException.Builder errorHandler = new IllegalAnnotationsException.Builder();
        builder.setErrorHandler(errorHandler);
        for (Class c : this.classes) {
            if (c != CompositeStructure.class) {
                builder.getTypeInfo(new Ref(c));
            }
        }
        this.hasSwaRef |= builder.hasSwaRef;
        RuntimeTypeInfoSet r2 = builder.link();
        errorHandler.check();
        assert r2 != null : "if no error was reported, the link must be a success";
        this.typeInfoSetCache = new WeakReference(r2);
        return r2;
    }
    
    public ElementBeanInfoImpl getElement(Class scope, QName name) {
        Map<QName, ElementBeanInfoImpl> m = (Map)this.elements.get(scope);
        if (m != null) {
            ElementBeanInfoImpl bi = (ElementBeanInfoImpl)m.get(name);
            if (bi != null) {
                return bi;
            }
        }
        m = (Map)this.elements.get(null);
        return (ElementBeanInfoImpl)m.get(name);
    }
    
    private ElementBeanInfoImpl getOrCreate(RuntimeElementInfo rei) {
        JaxBeanInfo<Object> bi = (JaxBeanInfo)this.beanInfos.get(rei);
        if (bi != null) {
            return (ElementBeanInfoImpl)bi;
        }
        return new ElementBeanInfoImpl(this, rei);
    }
    
    protected JaxBeanInfo getOrCreate(RuntimeEnumLeafInfo eli) {
        JaxBeanInfo<Object> bi = (JaxBeanInfo)this.beanInfos.get(eli);
        if (bi != null) {
            return bi;
        }
        bi = new LeafBeanInfoImpl(this, eli);
        this.beanInfoMap.put(bi.jaxbType, bi);
        return bi;
    }
    
    protected ClassBeanInfoImpl getOrCreate(RuntimeClassInfo ci) {
        ClassBeanInfoImpl bi = (ClassBeanInfoImpl)this.beanInfos.get(ci);
        if (bi != null) {
            return bi;
        }
        bi = new ClassBeanInfoImpl(this, ci);
        this.beanInfoMap.put(bi.jaxbType, bi);
        return bi;
    }
    
    protected JaxBeanInfo getOrCreate(RuntimeArrayInfo ai) {
        JaxBeanInfo<Object> abi = (JaxBeanInfo)this.beanInfos.get(ai);
        if (abi != null) {
            return abi;
        }
        abi = new ArrayBeanInfoImpl(this, ai);
        this.beanInfoMap.put(ai.getType(), abi);
        return abi;
    }
    
    public JaxBeanInfo getOrCreate(RuntimeTypeInfo e) {
        if ((e instanceof RuntimeElementInfo)) {
            return this.getOrCreate((RuntimeElementInfo)e);
        }
        if ((e instanceof RuntimeClassInfo)) {
            return this.getOrCreate((RuntimeClassInfo)e);
        }
        if ((e instanceof RuntimeLeafInfo)) {
            JaxBeanInfo<Object> bi = (JaxBeanInfo)this.beanInfos.get(e);
            assert bi != null;
            return bi;
        }
        else {
            if ((e instanceof RuntimeArrayInfo)) {
                return this.getOrCreate((RuntimeArrayInfo)e);
            }
            if (e.getType() == Object.class) {
                JaxBeanInfo<Object> bi = (JaxBeanInfo)this.beanInfoMap.get(Object.class);
                if (bi == null) {
                    bi = new AnyTypeBeanInfo(this, e);
                    this.beanInfoMap.put(Object.class, bi);
                    return bi;
                }
                return bi;
            }
            else {
                throw new IllegalArgumentException();
            }
        }
    }
    
    public final JaxBeanInfo getBeanInfo(Object o) {
        Class<Object> c = o.getClass();
        while (c != Object.class) {
            JaxBeanInfo<Object> bi = (JaxBeanInfo)this.beanInfoMap.get(c);
            if (bi != null) {
                return bi;
            }
            c = c.getSuperclass();
        }
        if ((o instanceof Element)) {
            return (JaxBeanInfo)this.beanInfoMap.get(Object.class);
        }
        Class[] interfaces = o.getClass().getInterfaces();
        int length = interfaces.length;
        int i = 0;
        while (i < length) {
            Class<Object> c2 = interfaces[i];
            JaxBeanInfo<Object> bi2 = (JaxBeanInfo)this.beanInfoMap.get(c2);
            if (bi2 != null) {
                return bi2;
            }
            ++i;
        }
        return null;
    }
    
    public final JaxBeanInfo getBeanInfo(Object o, boolean fatal) throws JAXBException {
        JaxBeanInfo<Object> bi = this.getBeanInfo(o);
        if (bi != null) {
            return bi;
        }
        if (fatal) {
            if ((o instanceof Document)) {
                throw new JAXBException(Messages.ELEMENT_NEEDED_BUT_FOUND_DOCUMENT.format(new Object[] { o.getClass() }));
            }
            throw new JAXBException(Messages.UNKNOWN_CLASS.format(new Object[] { o.getClass() }));
        }
        else {
            return null;
        }
    }
    
    public final <T> JaxBeanInfo<T> getBeanInfo(Class<T> clazz) {
        return (JaxBeanInfo)this.beanInfoMap.get(clazz);
    }
    
    public final <T> JaxBeanInfo<T> getBeanInfo(Class<T> clazz, boolean fatal) throws JAXBException {
        JaxBeanInfo<T> bi = this.getBeanInfo(clazz);
        if (bi != null) {
            return bi;
        }
        if (fatal) {
            throw new JAXBException(clazz.getName() + " is not known to this context");
        }
        return null;
    }
    
    public final Loader selectRootLoader(UnmarshallingContext.State state, TagName tag) {
        JaxBeanInfo<Object> beanInfo = (JaxBeanInfo)this.rootMap.get(tag.uri, tag.local);
        if (beanInfo == null) {
            return null;
        }
        return beanInfo.getLoader(this, true);
    }
    
    public JaxBeanInfo getGlobalType(QName name) {
        return (JaxBeanInfo)this.typeMap.get(name);
    }
    
    public String getNearestTypeName(QName name) {
        String[] all = new String[this.typeMap.size()];
        int i = 0;
        for (QName qn : this.typeMap.keySet()) {
            if (qn.getLocalPart().equals(name.getLocalPart())) {
                return qn.toString();
            }
            all[i++] = qn.toString();
        }
        String nearest = EditDistance.findNearest(name.toString(), all);
        if (EditDistance.editDistance(nearest, name.toString()) > 10) {
            return null;
        }
        return nearest;
    }
    
    public Set<QName> getValidRootNames() {
        Set<QName> r = new TreeSet(JAXBContextImpl.QNAME_COMPARATOR);
        for (QNameMap.Entry e : this.rootMap.entrySet()) {
            r.add(e.createQName());
        }
        return r;
    }
    
    public synchronized Encoded[] getUTF8NameTable() {
        if (this.utf8nameTable == null) {
            Encoded[] x = new Encoded[this.nameList.localNames.length];
            Encoded e;
            for (int i = 0; i < x.length; x[i] = e, ++i) {
                e = new Encoded(this.nameList.localNames[i]);
                e.compact();
            }
            this.utf8nameTable = x;
        }
        return this.utf8nameTable;
    }
    
    public int getNumberOfLocalNames() {
        return this.nameList.localNames.length;
    }
    
    public int getNumberOfElementNames() {
        return this.nameList.numberOfElementNames;
    }
    
    public int getNumberOfAttributeNames() {
        return this.nameList.numberOfAttributeNames;
    }
    
    static Transformer createTransformer(boolean disableSecureProcessing) {
        try {
            SAXTransformerFactory tf = (SAXTransformerFactory)XmlFactory.createTransformerFactory(disableSecureProcessing);
            return tf.newTransformer();
        }
        catch (TransformerConfigurationException e) {
            throw new Error(e);
        }
    }
    
    public static TransformerHandler createTransformerHandler(boolean disableSecureProcessing) {
        try {
            SAXTransformerFactory tf = (SAXTransformerFactory)XmlFactory.createTransformerFactory(disableSecureProcessing);
            return tf.newTransformerHandler();
        }
        catch (TransformerConfigurationException e) {
            throw new Error(e);
        }
    }
    
    static Document createDom(boolean disableSecurityProcessing) {
        synchronized (JAXBContextImpl.class) {
            if (JAXBContextImpl.db == null) {
                try {
                    DocumentBuilderFactory dbf = XmlFactory.createDocumentBuilderFactory(disableSecurityProcessing);
                    JAXBContextImpl.db = dbf.newDocumentBuilder();
                }
                catch (ParserConfigurationException e) {
                    throw new FactoryConfigurationError(e);
                }
            }
            return JAXBContextImpl.db.newDocument();
        }
    }
    
    public MarshallerImpl createMarshaller() {
        return new MarshallerImpl(this, null);
    }
    
    public UnmarshallerImpl createUnmarshaller() {
        return new UnmarshallerImpl(this, null);
    }
    
    public Validator createValidator() {
        throw new UnsupportedOperationException(Messages.NOT_IMPLEMENTED_IN_2_0.format(new Object[0]));
    }
    
    public JAXBIntrospector createJAXBIntrospector() {
        return new JAXBIntrospector() {
            public boolean isElement(Object object) {
                return this.getElementName(object) != null;
            }
            
            public QName getElementName(Object jaxbElement) {
                try {
                    return JAXBContextImpl.this.getElementName(jaxbElement);
                }
                catch (JAXBException e) {
                    return null;
                }
            }
        };
    }
    
    private NonElement<Type, Class> getXmlType(RuntimeTypeInfoSet tis, TypeReference tr) {
        if (tr == null) {
            throw new IllegalArgumentException();
        }
        XmlJavaTypeAdapter xjta = (XmlJavaTypeAdapter)tr.get(XmlJavaTypeAdapter.class);
        XmlList xl = (XmlList)tr.get(XmlList.class);
        Ref<Type, Class> ref = new Ref(this.annotationReader, tis.getNavigator(), tr.type, xjta, xl);
        return tis.getTypeInfo(ref);
    }
    
    public void generateEpisode(Result output) {
        if (output == null) {
            throw new IllegalArgumentException();
        }
        this.createSchemaGenerator().writeEpisodeFile(ResultFactory.createSerializer(output));
        return;
    }
    
    public void generateSchema(SchemaOutputResolver outputResolver) throws IOException {
        if (outputResolver == null) {
            throw new IOException(Messages.NULL_OUTPUT_RESOLVER.format(new Object[0]));
        }
        final SAXParseException[] e = { null };
        final SAXParseException[] w = { null };
        this.createSchemaGenerator().write(outputResolver, new ErrorListener() {
            public void error(SAXParseException exception) {
                e[0] = exception;
            }
            
            public void fatalError(SAXParseException exception) {
                e[0] = exception;
            }
            
            public void warning(SAXParseException exception) {
                w[0] = exception;
            }
            
            public void info(SAXParseException exception) {
            }
        });
        if (e[0] != null) {
            IOException x = new IOException(Messages.FAILED_TO_GENERATE_SCHEMA.format(new Object[0]));
            x.initCause(e[0]);
            throw x;
        }
        if (w[0] != null) {
            IOException x = new IOException(Messages.ERROR_PROCESSING_SCHEMA.format(new Object[0]));
            x.initCause(w[0]);
            throw x;
        }
        return;
    }
    
    private XmlSchemaGenerator<Type, Class, Field, Method> createSchemaGenerator() {
        RuntimeTypeInfoSet tis;
        try {
            tis = this.getTypeInfoSet();
        }
        catch (IllegalAnnotationsException e) {
            throw new AssertionError(e);
        }
        XmlSchemaGenerator<Type, Class, Field, Method> xsdgen = new XmlSchemaGenerator(tis.getNavigator(), tis);
        Set<QName> rootTagNames = new HashSet();
        for (RuntimeElementInfo ei : tis.getAllElements()) {
            rootTagNames.add(ei.getElementName());
        }
        for (RuntimeClassInfo ci : tis.beans().values()) {
            if (ci.isElement()) {
                rootTagNames.add(ci.asElement().getElementName());
            }
            else {
                continue;
            }
        }
        for (TypeReference tr : this.bridges.keySet()) {
            if (rootTagNames.contains(tr.tagName)) {
                continue;
            }
            if (tr.type == Void.TYPE || tr.type == Void.class) {
                xsdgen.add(tr.tagName, false, null);
            }
            else {
                if (tr.type == CompositeStructure.class) {
                    continue;
                }
                NonElement<Type, Class> typeInfo = this.getXmlType(tis, tr);
                xsdgen.add(tr.tagName, !tis.getNavigator().isPrimitive(tr.type), typeInfo);
            }
        }
        return xsdgen;
    }
    
    public QName getTypeName(TypeReference tr) {
        try {
            NonElement<Type, Class> xt = this.getXmlType(this.getTypeInfoSet(), tr);
            if (xt == null) {
                throw new IllegalArgumentException();
            }
            return xt.getTypeName();
        }
        catch (IllegalAnnotationsException e) {
            throw new AssertionError(e);
        }
    }
    
    public <T> Binder<T> createBinder(Class<T> domType) {
        if (domType == Node.class) {
            return this.createBinder();
        }
        return super.createBinder(domType);
    }
    
    public Binder<Node> createBinder() {
        return new BinderImpl(this, new DOMScanner());
    }
    
    public QName getElementName(Object o) throws JAXBException {
        JaxBeanInfo<Object> bi = this.getBeanInfo(o, true);
        if (!bi.isElement()) {
            return null;
        }
        return new QName(bi.getElementNamespaceURI(o), bi.getElementLocalName(o));
    }
    
    public QName getElementName(Class o) throws JAXBException {
        JaxBeanInfo<Class> bi = this.getBeanInfo(o, true);
        if (!bi.isElement()) {
            return null;
        }
        return new QName(bi.getElementNamespaceURI(o), bi.getElementLocalName(o));
    }
    
    public Bridge createBridge(TypeReference ref) {
        return (Bridge)this.bridges.get(ref);
    }
    
    @NotNull
    public BridgeContext createBridgeContext() {
        return new BridgeContextImpl(this);
    }
    
    public RawAccessor getElementPropertyAccessor(Class wrapperBean, String nsUri, String localName) throws JAXBException {
        JaxBeanInfo<Object> bi = this.getBeanInfo(wrapperBean, true);
        if (!(bi instanceof ClassBeanInfoImpl)) {
            throw new JAXBException(wrapperBean + " is not a bean");
        }
        for (ClassBeanInfoImpl cb = (ClassBeanInfoImpl)bi; cb != null; cb = cb.superClazz) {
            Property[] properties = cb.properties;
            int length = properties.length;
            int i = 0;
            while (i < length) {
                Property p = properties[i];
                final Accessor acc = p.getElementPropertyAccessor(nsUri, localName);
                if (acc != null) {
                    return new RawAccessor() {
                        public Object get(Object bean) throws AccessorException {
                            return acc.getUnadapted(bean);
                        }
                        
                        public void set(Object bean, Object value) throws AccessorException {
                            acc.setUnadapted(bean, value);
                        }
                    };
                }
                ++i;
            }
        }
        throw new JAXBException(new QName(nsUri, localName) + " is not a valid property on " + wrapperBean);
    }
    
    public List<String> getKnownNamespaceURIs() {
        return Arrays.asList(this.nameList.namespaceURIs);
    }
    
    public String getBuildId() {
        Package pkg = this.getClass().getPackage();
        if (pkg == null) {
            return null;
        }
        return pkg.getImplementationVersion();
    }
    
    public String toString() {
        StringBuilder buf = new StringBuilder(Which.which(this.getClass()) + " Build-Id: " + this.getBuildId());
        buf.append("\nClasses known to this context:\n");
        Set<String> names = new TreeSet();
        for (Class key : this.beanInfoMap.keySet()) {
            names.add(key.getName());
        }
        for (String name : names) {
            buf.append("  ").append(name).append('\n');
        }
        return buf.toString();
    }
    
    public String getXMIMEContentType(Object o) {
        JaxBeanInfo<Object> bi = this.getBeanInfo(o);
        if (!(bi instanceof ClassBeanInfoImpl)) {
            return null;
        }
        ClassBeanInfoImpl<Object> cb = (ClassBeanInfoImpl)bi;
        for (Property p : cb.properties) {
            if ((p instanceof AttributeProperty)) {
                AttributeProperty ap = (AttributeProperty)p;
                if (ap.attName.equals("http://www.w3.org/2005/05/xmlmime", "contentType")) {
                    try {
                        return (String)ap.xacc.print(o);
                    }
                    catch (AccessorException e) {
                        return null;
                    }
                    catch (SAXException e2) {
                        return null;
                    }
                    catch (ClassCastException e3) {
                        return null;
                    }
                }
            }
        }
        return null;
    }
    
    public JAXBContextImpl createAugmented(Class<?> clazz) throws JAXBException {
        Class[] newList = new Class[this.classes.length + 1];
        System.arraycopy(this.classes, 0, newList, 0, this.classes.length);
        newList[this.classes.length] = clazz;
        JAXBContextImpl.JAXBContextBuilder builder = new JAXBContextImpl.JAXBContextBuilder(this);
        builder.setClasses(newList);
        return builder.build();
    }
    
    JAXBContextImpl(JAXBContextImpl.JAXBContextBuilder x0, JAXBContextImpl.1 x1) throws JAXBException {
        this(x0);
    }
    
    static {
        QNAME_COMPARATOR = new Comparator<QName>() {
            @Override
            public int compare(QName lhs, QName rhs) {
                int r = lhs.getLocalPart().compareTo(rhs.getLocalPart());
                if (r != 0) {
                    return r;
                }
                return lhs.getNamespaceURI().compareTo(rhs.getNamespaceURI());
            }
        };
    }
    
    public static class JAXBContextBuilder
    {
        private boolean retainPropertyInfo;
        private boolean supressAccessorWarnings;
        private String defaultNsUri;
        @NotNull
        private RuntimeAnnotationReader annotationReader;
        @NotNull
        private Map<Class, Class> subclassReplacements;
        private boolean c14nSupport;
        private Class[] classes;
        private Collection<TypeReference> typeRefs;
        private boolean xmlAccessorFactorySupport;
        private boolean allNillable;
        private boolean improvedXsiTypeHandling;
        private boolean disableSecurityProcessing;
        
        public JAXBContextBuilder() {
            this.retainPropertyInfo = false;
            this.supressAccessorWarnings = false;
            this.defaultNsUri = "";
            this.annotationReader = new RuntimeInlineAnnotationReader();
            this.subclassReplacements = Collections.emptyMap();
            this.c14nSupport = false;
            this.xmlAccessorFactorySupport = false;
            this.improvedXsiTypeHandling = true;
            this.disableSecurityProcessing = true;
        }
        
        public JAXBContextBuilder(JAXBContextImpl baseImpl) {
            this.retainPropertyInfo = false;
            this.supressAccessorWarnings = false;
            this.defaultNsUri = "";
            this.annotationReader = new RuntimeInlineAnnotationReader();
            this.subclassReplacements = Collections.emptyMap();
            this.c14nSupport = false;
            this.xmlAccessorFactorySupport = false;
            this.improvedXsiTypeHandling = true;
            this.disableSecurityProcessing = true;
            this.supressAccessorWarnings = baseImpl.supressAccessorWarnings;
            this.retainPropertyInfo = baseImpl.retainPropertyInfo;
            this.defaultNsUri = baseImpl.defaultNsUri;
            this.annotationReader = baseImpl.annotationReader;
            this.subclassReplacements = baseImpl.subclassReplacements;
            this.c14nSupport = baseImpl.c14nSupport;
            this.classes = baseImpl.classes;
            this.typeRefs = baseImpl.bridges.keySet();
            this.xmlAccessorFactorySupport = baseImpl.xmlAccessorFactorySupport;
            this.allNillable = baseImpl.allNillable;
            this.disableSecurityProcessing = baseImpl.disableSecurityProcessing;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setRetainPropertyInfo(boolean val) {
            this.retainPropertyInfo = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setSupressAccessorWarnings(boolean val) {
            this.supressAccessorWarnings = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setC14NSupport(boolean val) {
            this.c14nSupport = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setXmlAccessorFactorySupport(boolean val) {
            this.xmlAccessorFactorySupport = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setDefaultNsUri(String val) {
            this.defaultNsUri = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setAllNillable(boolean val) {
            this.allNillable = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setClasses(Class[] val) {
            this.classes = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setAnnotationReader(RuntimeAnnotationReader val) {
            this.annotationReader = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setSubclassReplacements(Map<Class, Class> val) {
            this.subclassReplacements = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setTypeRefs(Collection<TypeReference> val) {
            this.typeRefs = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setImprovedXsiTypeHandling(boolean val) {
            this.improvedXsiTypeHandling = val;
            return this;
        }
        
        public JAXBContextImpl.JAXBContextBuilder setDisableSecurityProcessing(boolean val) {
            this.disableSecurityProcessing = val;
            return this;
        }
        
        public JAXBContextImpl build() throws JAXBException {
            if (this.defaultNsUri == null) {
                this.defaultNsUri = "";
            }
            if (this.subclassReplacements == null) {
                this.subclassReplacements = Collections.emptyMap();
            }
            if (this.annotationReader == null) {
                this.annotationReader = new RuntimeInlineAnnotationReader();
            }
            if (this.typeRefs == null) {
                this.typeRefs = Collections.emptyList();
            }
            return new JAXBContextImpl(this);
        }
    }
}
